'''
Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
'''

class LcnAddr(object):
    """Represents a LCN address (module or group)
    If the segment id is 0, the address object points to modules/groups which are in the segment where the bus coupler is
    connected to. This is also the case if no segment coupler is present at all.

    :param    int    seg_id:    Segment id
                                (0 = Local,
                                1..2 = Not allowed (but "seen in the wild")
                                3 = Broadcast,
                                4 = Status messages,
                                5..127,
                                128 = Segment-bus disabled (valid value))
    :param    bool   is_group:  Indicates whether address point to a module (False) or a group (True)


    If address represents a **module**:
    
    :param    int    id:        Module id
                                (1 = LCN-PRO,
                                2 = LCN-GVS/LCN-W,
                                4 = PCHK,
                                5..254,
                                255 = Unprog. (valid, but irrelevant here))

    
    If address represents a **group**:

    :param    int    id:        Group id
                                (3 = Broadcast,
                                4 = Status messages,
                                5..254)
    """
    def __init__(self, seg_id = -1, id = -1, is_group = False):
        """Constructor
        """
        self.seg_id = seg_id
        self.id = id
        self._is_group = is_group
    
    def is_group(self):
        """Gets the address' module or group id (discarding the concrete type).
        
        :return:    Returns whether address points to a module(False) or group(True)
        :rtype:     bool
        """
        return self._is_group

    def get_seg_id(self):
        """Gets the logical segment id.
        
        :return:    The (logical) segment id
        :rtype:     int
        """
        return self.seg_id
    
    def get_physical_seg_id(self, local_seg_id):
        """Gets the physical segment id ("local" segment replaced with 0).
        Can be used to send data into the LCN bus.

        :param    int    local_seg_id:    The segment id of the local segment
        
        :return:    The physical segment id
        :rtype:     int
        """
        return 0 if (self.seg_id == local_seg_id) else self.seg_id

    def get_id(self):
        """
        Gets the module id.

        :return:    The module id
        :rtype:     int
        """
        return self.id
 
    def is_valid(self):
        """
        Returns if the current address is valid.

        :return:    True, if address is a valid group/module address, otherwise False
        :rtype:     bool
        """
        if self.is_group():
            """
            seg_id:
            0 = Local, 1..2 = Not allowed (but "seen in the wild")
            3 = Broadcast, 4 = Status messages, 5..127, 128 = Segment-bus disabled (valid value)
            id:
            3 = Broadcast, 4 = Status messages, 5..254
            """
            return (self.seg_id >= 0) & (self.seg_id <= 128) & (self.id >= 3) & (self.id < 254)
        else:
            """
            seg_id:
            0 = Local, 1..2 = Not allowed (but "seen in the wild")
            3 = Broadcast, 4 = Status messages, 5..127, 128 = Segment-bus disabled (valid value)
            id:
            1 = LCN-PRO, 2 = LCN-GVS/LCN-W, 4 = PCHK, 5..254, 255 = Unprog. (valid, but irrelevant here)
            """
            return (self.seg_id >= 0) & (self.seg_id <= 128) & (self.id >= 1) & (self.id < 254)
    
    def __hash__(self):
        if self.is_valid():
            return (self.is_group() << 9) + (reverse_uint8(self.get_id()) << 8) + (reverse_uint8(self.get_seg_id()))
        else:
            return -1
    
    def __eq__(self, obj):
        return (self.is_group() == obj.is_group()) & (self.get_seg_id() == obj.get_seg_id()) & (self.get_id() == obj.get_id())



# only execute, if not defined before
if not 'reversed_uint8' in dir():
    reversed_uint8 = [0]*256
    for i in range(256):
        reversed = 0
        for j in range(8):
            if ((i & (1 << j)) != 0):
                reversed |= (0x80 >> j)
        reversed_uint8[i] = reversed


def reverse_uint8(value):
    if (value < 0 | value > 255):
        raise ValueError('Invalid value.')
    return reversed_uint8[value]




'''
Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
'''

import logging
import re

from pypck import lcn_defs
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckParser, PckGenerator
from pypck.timeout_retry import DEFAULT_TIMEOUT_MSEC

_LOGGER = logging.getLogger(__name__)


class Input(object):
    """
    Parent class for all input data read from LCN-PCHK.
    
    An implementation of :class:`~pypck.input.Input` has to provide easy accessible attributes
    and/or methods to expose the PCK command properties to the user.
    Each Input object provides an implementation of :func:`~pypck.input.Input.try_parse`
    static method, to parse the given plain text PCK command. If the command can be parsed by
    the Input object, a list of instances of :class:`~pypck.input.Input` is returned. 
    If parsing is successful, the :func:`~pypck.input.Input.process` method can be called
    to trigger further actions.
    """

    def __init__(self):
        pass

    @staticmethod
    def try_parse(input):
        """
        Tries to parse the given input text.
        Will return a list of parsed Inputs. The list might be empty (but not null).
       
        :param    input:     the input data received from LCN-PCHK
        :type     input:     :class:`~pypck.input.Input`
        :return:             The parsed Inputs (never null)
        :rtype:              List with instances of :class:`~pypck.input.Input`
        """
        raise NotImplementedError

    def process(self, conn):
        """
        Processes the :class:`~pypck.input.Input` instance and triggers further actions.
        
        :param ~pypck.connection.PchkConnectionManager conn: Connection manager object 
        """
        raise NotImplementedError


class ModInput(Input):
    """
    Parent class of all inputs having an LCN module as its source.
    
    The class in inherited from :class:`~pypck.input.Input`
    """

    def __init__(self, physical_source_addr):
        super().__init__()
        self.physical_source_addr = physical_source_addr
        self.logical_source_addr = LcnAddr()

    def get_logical_source_addr(self):
        return self.logical_source_addr

    def process(self, conn):
        if conn.is_ready():  # Skip if we don't have all necessary bus info yet
            self.logical_source_addr = conn.physical_to_logical(self.physical_source_addr)

# ## Plain text inputs


class AuthOk(Input):

    @staticmethod
    def try_parse(input):
        if input == PckParser.AUTH_OK:
            return [AuthOk()]

    def process(self, conn):
        conn.on_auth_ok()


class AuthPassword(Input):

    @staticmethod
    def try_parse(input):
        if input == PckParser.AUTH_PASSWORD:
            return [AuthPassword()]

    def process(self, conn):
        conn.send_command(conn.password)


class AuthUsername(Input):

    @staticmethod
    def try_parse(input):
        if input == PckParser.AUTH_USERNAME:
            return [AuthUsername()]

    def process(self, conn):
        conn.send_command(conn.username)


class LcnConnState(Input):

    def __init__(self, is_lcn_connected):
        super().__init__()
        self._is_lcn_connected = is_lcn_connected

    @property
    def is_lcn_connected(self):
        return self._is_lcn_connected

    @staticmethod
    def try_parse(input):
        if input == PckParser.LCNCONNSTATE_CONNECTED:
            return [LcnConnState(True)]
        elif input == PckParser.LCNCONNSTATE_DISCONNECTED:
            return [LcnConnState(False)]

    def process(self, conn):
        if self.is_lcn_connected:
            _LOGGER.debug('{}: LCN is connected.'.format(conn.connection_id))
            conn.on_successful_login()
            conn.send_command(PckGenerator.set_operation_mode(conn.dim_mode, conn.status_mode))
        else:
            _LOGGER.debug('{}: LCN is not connected.'.format(conn.connection_id))

# ## Inputs received from modules


class ModAck(ModInput):

    def __init__(self, physical_source_addr, code):
        super().__init__(physical_source_addr)
        self.code = code

    def get_code(self):
        return self.code

    @staticmethod
    def try_parse(input):
        matcher_pos = PckParser.PATTERN_ACK_POS.match(input)
        if matcher_pos:
            addr = LcnAddr(int(matcher_pos.group('seg_id')),
                           int(matcher_pos.group('mod_id')))
            return [ModAck(addr, -1)]

        matcher_neg = PckParser.PATTERN_ACK_NEG.match(input)
        if matcher_neg:
            addr = LcnAddr(int(matcher_neg.group('seg_id')),
                           int(matcher_neg.group('mod_id')))
            return [ModAck(addr, matcher_neg.group('code'))]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        module_conn.on_ack(self.code, DEFAULT_TIMEOUT_MSEC)


class ModSk(ModInput):
    """Segment information received from an LCN segment coupler.
    """

    def __init__(self, physical_source_addr, reported_seg_id):
        super().__init__(physical_source_addr)
        self.reported_seg_id = reported_seg_id

    def get_reported_seg_id(self):
        return self.reported_seg_id

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_SK_RESPONSE.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModSk(addr, int(matcher.group('id')))]

    def process(self, conn):
        if self.physical_source_addr.seg_id == 0:
            conn.set_local_seg_id(self.reported_seg_id)
        super().process(conn)  # Will replace source segment 0 with the local segment id
        conn.status_segment_scan.cancel()
        for module_conn in conn.address_conns.values():
            conn.loop.create_task(module_conn.activate_status_request_handlers())


class ModSn(ModInput):
    """Serial number and firmware version received from an LCN module.
    """

    def __init__(self, physical_source_addr, sw_age):
        super().__init__(physical_source_addr)
        self.sw_age = sw_age

    def get_sw_age(self):
        return self.sw_age

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_SN.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModSn(addr, int(matcher.group('sw_age'), 16))]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        module_conn.set_sw_age(self.sw_age)
        module_conn.request_sw_age.cancel()


class ModStatusOutput(ModInput):
    """Status of an output-port received from an LCN module.
    """

    def __init__(self, physical_source_addr, output_id, percent):
        super().__init__(physical_source_addr)
        self.output_id = output_id
        self.percent = percent

    def get_output_id(self):
        return self.output_id

    def get_percent(self):
        return self.percent

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_OUTPUT_PERCENT.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModStatusOutput(addr, int(matcher.group('output_id')) - 1, float(matcher.group('percent')))]

        matcher = PckParser.PATTERN_STATUS_OUTPUT_NATIVE.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModStatusOutput(addr, int(matcher.group('output_id')), float(matcher.group('value')) / 2.)]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        # module_conn.request_status_outputs[self.output_id].cancel()
        module_conn.new_input(self)


class ModStatusRelays(ModInput):
    """Status of 8 relays received from an LCN module.
    """

    def __init__(self, physical_source_addr, states):
        super().__init__(physical_source_addr)
        self.states = states

    def get_state(self, relay_id):
        """
        Gets the state of a single relay.
        
        :param    int    relay_id:    Relay id (0..7)
        
        :return:                      The relay's state
        :rtype:   bool
        """
        return self.states[relay_id]

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_RELAYS.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModStatusRelays(addr, PckParser.get_boolean_value(int(matcher.group('byte_value'))))]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        # module_conn.request_status_relays.cancel()
        module_conn.new_input(self)


class ModStatusBinSensors(ModInput):
    """Status of 8 binary sensors received from an LCN module.
    """

    def __init__(self, physical_source_addr, states):
        super().__init__(physical_source_addr)
        self.states = states

    def get_state(self, bin_sensor_id):
        """
        Gets the state of a single binary-sensor.
        
        :param    int    bin_sensor_id:    Binary sensor id (0..7)
        
        :return:                           The binary-sensor's state
        :rtype:   bool
        """
        return not self.states[bin_sensor_id]

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_BINSENSORS.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            return [ModStatusBinSensors(addr, PckParser.get_boolean_value(int(matcher.group('byte_value'))))]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        # module_conn.request_status_relays.cancel()
        module_conn.new_input(self)


class ModStatusVar(ModInput):
    """
    Status of a variable received from an LCN module.
    """

    def __init__(self, physical_source_addr, orig_var, value):
        super().__init__(physical_source_addr)
        self.orig_var = orig_var
        self.value = value
        self.var = self.orig_var

    def get_var(self):
        """
        Gets the variable's real type.
        
        :return:        The real type
        :rtype:        :class:`~pypck.lcn_defs.Var`
        """
        return self.var

    def get_value(self):
        """
        Gets the variable's value.

        :return:    The value of the variable.
        :rtype:     int
        """
        return self.value

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_VAR.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            var = lcn_defs.Var.var_id_to_var(int(matcher.group('id')) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group('value')))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_SETVAR.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            var = lcn_defs.Var.set_point_id_to_var(int(matcher.group('id')) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group('value')))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_THRS.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            var = lcn_defs.Var.thrs_id_to_var(int(matcher.group('register_id')) - 1, int(matcher.group('thrs_id')) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group('value')))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_STATUS_S0INPUT.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            var = lcn_defs.Var.s0_id_to_var(int(matcher.group('id')) - 1)
            value = lcn_defs.VarValue.from_native(int(matcher.group('value')))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_VAR_GENERIC.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            var = lcn_defs.Var.UNKNOWN
            value = lcn_defs.VarValue.from_native(int(matcher.group('value')))
            return [ModStatusVar(addr, var, value)]

        matcher = PckParser.PATTERN_THRS5.match(input)
        if matcher:
            ret = []
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            for thrs_id in range(5):
                var = lcn_defs.Var.var_id_to_var(int(matcher.group('id')) - 1)
                value = lcn_defs.VarValue.from_native(int(matcher.group('value{:d}'.format(thrs_id + 1))))
                ret.append(ModStatusVar(addr, var, value))
            return ret

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        if self.orig_var == lcn_defs.Var.UNKNOWN:
            self.var = module_conn.get_last_requested_var_without_type_in_response()
        else:
            self.var = self.orig_var

        if self.var != lcn_defs.Var.UNKNOWN:
            if module_conn.get_last_requested_var_without_type_in_response() == self.var:
                module_conn.set_last_requested_var_without_type_in_response(lcn_defs.Var.UNKNOWN)  # Reset
#             if self.var in module_conn.request_status_vars:
#                 module_conn.request_status_vars[self.var].cancel()
        module_conn.new_input(self)


class ModStatusLedsAndLogicOps(ModInput):
    """Status of LEDs and logic-operations received from an LCN module.

    :param    int      physicalSourceAddr:   The physical source address
    :param    states_led:                    The 12 LED states
    :type     states_led:                    list(:class:`~pypck.lcn_defs.LedStatus`)
    :param    states_logic_ops:              The 4 logic-operation states
    :type     states_logic_ops:              list(:class:`~pypck.lcn_defs.LogicOpStatus`)         
    """

    def __init__(self, physical_source_addr, states_led, states_logic_ops):
        """Constructor."""
        super().__init__(physical_source_addr)
        self.states_led = states_led  # 12x LED status.
        self.states_logic_ops = states_logic_ops  # 4x logic-operation status.

    def get_led_state(self, led_id):
        """Gets the status of a single LED.

        :param    int    led_id:   LED id (0..11)
        :return:                   The LED's status
        :rtype:   list(:class:`~pypck.lcn_defs.LedStatus`)
        """
        return self.states_led[led_id]

    def get_logic_op_state(self, logic_op_id):
        """Gets the status of a single logic operation.

        :param    int    logic_op_id:    Logic operation id (0..3)
        :return:    The logic-operation's status
        :rtype:     list(:class:`~pypck.lcn_defs.LogicOpStatus`)
        """
        return self.states_logic_ops[logic_op_id]

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_LEDSANDLOGICOPS.match(input)
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))

            led_states = matcher.group('led_states').upper()
            states_leds = [lcn_defs.LedStatus(led_state) for led_state in led_states]

            logic_op_states = matcher.group('logic_op_states').upper()
            states_logic_ops = [lcn_defs.LogicOpStatus(logic_op_state) for logic_op_state in logic_op_states]
            return [ModStatusLedsAndLogicOps(addr, states_leds, states_logic_ops)]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        # module_conn.request_status_leds_and_logic_ops.cancel()
        module_conn.new_input(self)


class ModStatusKeyLocks(ModInput):
    """
    Status of locked keys received from an LCN module.
    
    :param    int                physicalSourceAddr:    The source address
    :param    list(list(bool))   states:                The 4x8 key-lock states        
    """

    def __init__(self, physical_source_id, states):
        """Constructor.
        """
        super().__init__(physical_source_id)
        self.states = states

    def get_state(self, table_id, key_id):
        """Gets the lock-state of a single key.

        :param    int    tableId:    Table id: (0..3  =>  A..D)
        :param    int    keyId:      Key id (0..7  =>  1..8)
        :return:  The key's lock-state
        :rtype:   bool
        """
        return self.states[table_id][key_id]

    @staticmethod
    def try_parse(input):
        matcher = PckParser.PATTERN_STATUS_KEYLOCKS.match(input)
        states = []
        if matcher:
            addr = LcnAddr(int(matcher.group('seg_id')),
                           int(matcher.group('mod_id')))
            for i in range(4):
                s = matcher.group('table{:d}'.format(i))
                if s is not None:
                    states.append(PckParser.get_boolean_value(int(s)))
#                else:
#                     states.append([])
            return [ModStatusKeyLocks(addr, states)]

    def process(self, conn):
        super().process(conn)  # Will replace source segment 0 with the local segment id
        module_conn = conn.get_address_conn(self.logical_source_addr)
        # module_conn.request_status_key_locks.cancel()
        module_conn.new_input(self)

# ## Other inputs


class Unknown(Input):

    def __init__(self, input):
        super().__init__()
        self._input = input

    @staticmethod
    def try_parse(input):
        return [Unknown(input)]

    @property
    def input(self):
        return self._input

    def process(self, conn):
        pass


class InputParser(object):
    parsers = [AuthUsername,
               AuthPassword,
               AuthOk,
               LcnConnState,
               ModAck,
               ModSk,
               ModSn,
               ModStatusOutput,
               ModStatusRelays,
               ModStatusBinSensors,
               ModStatusVar,
               ModStatusLedsAndLogicOps,
               ModStatusKeyLocks,
               Unknown]

    @staticmethod
    def parse(input):
        for parser in InputParser.parsers:
            ret = parser.try_parse(input)
            if ret is not None:
                return ret




'''
Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
'''

import re
from pypck import lcn_defs


class PckParser(object):
    """Helpers to parse LCN-PCK commands.
    
    LCN-PCK is the command-syntax used by LCN-PCHK to send and receive LCN commands.
    """
    # Authentication at LCN-PCHK: Request user name.
    AUTH_USERNAME = 'Username:'

    # Authentication at LCN-PCHK: Request password.
    AUTH_PASSWORD = 'Password:'

    # Authentication at LCN-PCHK succeeded.
    AUTH_OK = 'OK';

    # LCN-PK/PKU is connected.
    LCNCONNSTATE_CONNECTED = '$io:#LCN:connected'

    # LCN-PK/PKU is disconnected.
    LCNCONNSTATE_DISCONNECTED = '$io:#LCN:disconnected'

    # Pattern to parse positive acknowledges.
    PATTERN_ACK_POS = re.compile(r'-M(?P<seg_id>\d{3})(?P<mod_id>\d{3})!')

    # Pattern to parse negative acknowledges.
    PATTERN_ACK_NEG = re.compile(r'-M(?P<seg_id>\d{3})(?P<mod_id>\d{3})(?P<code>\d+)')

    # Pattern to parse segment coupler responses.
    PATTERN_SK_RESPONSE = re.compile(r'=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SK(?P<id>\d+)')

    # Pattern to parse serial number and firmware date responses.
    PATTERN_SN = re.compile(r'=M(?P<seg_id>\d{3})(?P<mod_id>\d{3}).SN(?P<sn>[0-9|A-F]{10})(?P<manu>[0-9|A-F]{2})FW(?P<sw_age>[0-9|A-F]{6})HW(?P<hw_type>\d+)')

    # Pattern to parse output-port status responses in percent.
    PATTERN_STATUS_OUTPUT_PERCENT = re.compile(r':M(?P<seg_id>\d{3})(?P<mod_id>\d{3})A(?P<output_id>\d)(?P<percent>\d+)')

    # Pattern to parse output-port status responses in native format (0..200).
    PATTERN_STATUS_OUTPUT_NATIVE = re.compile(r':M(?P<seg_id>\d{3})(?P<mod_id>\d{3})O(?P<output_id>\d)(?P<value>\d+)')

    # Pattern to parse relays status responses.
    PATTERN_STATUS_RELAYS = re.compile(r':M(?P<seg_id>\d{3})(?P<mod_id>\d{3})Rx(?P<byte_value>\d+)')

    # Pattern to parse binary-sensors status responses.
    PATTERN_STATUS_BINSENSORS = re.compile(r':M(?P<seg_id>\d{3})(?P<mod_id>\d{3})Bx(?P<byte_value>\d+)')

    # Pattern to parse variable 1-12 status responses (since 170206).
    PATTERN_STATUS_VAR = re.compile(r'%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.A(?P<id>\d{3})(?P<value>\d+)')

    # Pattern to parse set-point variable status responses (since 170206).
    PATTERN_STATUS_SETVAR = re.compile(r'%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.S(?P<id>\d)(?P<value>\d+)')

    # Pattern to parse threshold status responses (since 170206).
    PATTERN_STATUS_THRS = re.compile(r'%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.T(?P<register_id>\d)(?P<thrs_id>\d)(?P<value>\d+)')

    # Pattern to parse S0-input status responses (LCN-BU4L).
    PATTERN_STATUS_S0INPUT = re.compile(r'%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.C(?P<id>\d)(?P<value>\d+)')

    # Pattern to parse generic variable status responses (concrete type unknown, before 170206).
    PATTERN_VAR_GENERIC = re.compile(r'%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.(?P<value>\d+)')

    # Pattern to parse threshold register 1 status responses (5 values, before 170206). */
    PATTERN_THRS5 = re.compile(r'=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.S1(?P<value1>\d{5})(?P<value2>\d{5})(?P<value3>\d{5})(?P<value4>\d{5})(?P<value5>\d{5})(?P<hyst>\d{5})')

    # Pattern to parse status of LEDs and logic-operations responses.
    PATTERN_STATUS_LEDSANDLOGICOPS = re.compile(r'=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.TL(?P<led_states>[AEBF]{12})(?P<logic_op_states>[NTV]{4})')

    # Pattern to parse key-locks status responses.
    PATTERN_STATUS_KEYLOCKS = re.compile(r'=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.TX(?P<table0>\d{3})(?P<table1>\d{3})(?P<table2>\d{3})((?P<table3>\d{3}))?')

    @staticmethod
    def get_boolean_value(input_byte):
        if (input_byte < 0) or (input_byte > 255):
            raise ValueError('Invalid input_byte.')

        result = []
        for i in range(8):
            result.append((input_byte & (1 << i)) != 0)
        return result


class PckGenerator(object):
    """Helpers to generate LCN-PCK commands.
    
    LCN-PCK is the command-syntax used by LCN-PCHK to send and receive LCN commands.
    """
    # Terminates a PCK command
    TERMINATION = '\n'

    TABLE_NAMES = ['A', 'B', 'C', 'D']

    @staticmethod
    def ping(counter):
        """Generates a keep-alive.
        LCN-PCHK will close the connection if it does not receive any commands from
        an open :class:`~pypck.connection.PchkConnectionManager` for a specific period (10 minutes by default).
        
        :param    int    counter:    The current ping's id (optional, but 'best practice'). Should start with 1
        :return:    The PCK command as text
        :rtype:    str
        """
        return '^ping{:d}'.format(counter)

    @staticmethod
    def set_operation_mode(dim_mode, status_mode):
        """Generates a PCK command that will set the LCN-PCHK connection's operation mode.
        This influences how output-port commands and status are interpreted and must be
        in sync with the LCN bus.
        
        :param    OuputPortDimMode        dimMode:        The dimming mode (50/200 steps)
        :param    OutputPortStatusMode    statusMode:     The status mode (percent/native)
        :return:    The PCK command as text
        :rtype:    str
        """
        return '!OM{:s}{:s}'.format('1' if (dim_mode == lcn_defs.OutputPortDimMode.STEPS200) else '0',
                                    'P' if (status_mode == lcn_defs.OutputPortStatusMode.PERCENT) else 'N')

    @staticmethod
    def generate_address_header(addr, local_seg_id, wants_ack):
        return '>{:s}{:03d}{:03d}{:s}'.format('G' if addr.is_group() else 'M',
                                              addr.get_physical_seg_id(local_seg_id),
                                              addr.get_id(),
                                              '!' if wants_ack else '.')

    @staticmethod
    def segment_coupler_scan():
        """Generates a scan-command for LCN segment-couplers.
        Used to detect the local segment (where the physical bus connection is located).

        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return 'SK'

    @staticmethod
    def request_sn():
        """Generates a firmware/serial-number request.
        
        :return: The PCK command (without address header) as text
        :rtype:    str
        """
        return 'SN'

    @staticmethod
    def request_output_status(output_id):
        """Generates an output-port status request.

        :param    int    output_id:    Output id 0..3
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError('Invalid output_id.')
        return 'SMA{:d}'.format(output_id + 1)

    @staticmethod
    def dim_ouput(output_id, percent, ramp):
        """Generates a dim command for a single output-port.

        :param    int    output_it:    Output id 0..3
        :param    int    percent:      Brightness in percent 0..100
        :param    int    ramp:         Ramp value
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError('Invalid output_id.')
        n = int(round(percent * 2))
        ramp = int(ramp)
        if (n % 2) == 0:  # Use the percent command (supported by all LCN-PCHK versions)
            return 'A{:d}DI{:03d}{:03d}'.format(output_id + 1, n // 2, ramp)
        else:  # We have a ".5" value. Use the native command (supported since LCN-PCHK 2.3)
            return 'O{:d}DI{:03d}{:03d}'.format(output_id + 1, n, ramp)

    @staticmethod
    def dim_all_outputs(percent, ramp, is1805=False):
        """Generates a dim command for all output-ports.

        :param    int    percent:    Brightness in percent 0..100
        :param    int    ramp:       Ramp value
        :param    bool   is1805:     True if the target module's firmware is 180501 or newer, otherwise False
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        n = round(percent * 2)
        if is1805:
            return 'OY{:03d}{:03d}{:03d}{:03d}{:03d}'.format(n, n, n, n, ramp)  # Supported since LCN-PCHK 2.61

        if n == 0:  # All off
            return 'AA{:03d}'.format(ramp)
        elif (n == 200):  # All on
            return 'AE{:03d}'.format(ramp)

        # This is our worst-case: No high-res, no ramp
        return 'AH{:03d}'.format(n / 2)

    @staticmethod
    def rel_output(output_id, percent):
        """Generates a command to change the value of an output-port.

        :param    int    output_id:    Output id 0..3
        :param    int    percent:      Relative percentage -100..100
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError('Invalid output_id.')

        n = round(percent * 2)
        if n % 2 == 0:  # Use the percent command (supported by all LCN-PCHK versions)
            return 'A{:d}{:s}{:03d}'.format(output_id + 1, 'AD' if percent >= 0 else 'SB', abs(n // 2))
        else:  # We have a ".5" value. Use the native command (supported since LCN-PCHK 2.3)
            return 'O{:d}{:s}{:03d}'.format(output_id + 1, 'AD' if percent >= 0 else 'SB', abs(n))

    @staticmethod
    def toggle_output(output_id, ramp):
        """Generates a command that toggles a single output-port (on->off, off->on).

        :param    int    output_id:    Output id 0..3
        :param    int    ramp:         Ramp value
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError('Invalid output_id.')
        return 'A{:d}TA{:03d}'.format(output_id + 1, ramp)

    @staticmethod
    def toggle_all_outputs(ramp):
        """Generates a command that toggles all output-ports (on->off, off->on).

        :param    int    ramp:    Ramp value
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return 'AU{:03d}'.format(ramp)

    @staticmethod
    def request_relays_status():
        """Generates a relays-status request.
         
        :return: The PCK command (without address header) as text
        :rtype:    str
        """
        return 'SMR'

    @staticmethod
    def control_relays(states):
        """Generates a command to control relays.

        :param     RelayStateModifier    states:    The 8 modifiers for the relay states as a list
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if len(states) != 8:
            raise ValueError('Invalid states length.')
        ret = 'R8'
        for state in states:
            ret += state.value

        return ret

    @staticmethod
    def control_motors(states):
        """Generates a command to control motors via relays.
        
        :param    MotorStateModifier    states:    The 4 modifiers for the motor states as a list
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if len(states) != 4:
            raise ValueError('Invalid states length.')
        ret = 'R8'
        for state in states:
            if state == lcn_defs.MotorStateModifier.UP:
                ret += lcn_defs.RelayStateModifier.ON.value
                ret += lcn_defs.RelayStateModifier.OFF.value
            elif state == lcn_defs.MotorStateModifier.DOWN:
                ret += lcn_defs.RelayStateModifier.ON.value
                ret += lcn_defs.RelayStateModifier.ON.value
            elif state == lcn_defs.MotorStateModifier.STOP:
                ret += lcn_defs.RelayStateModifier.OFF.value
                ret += lcn_defs.RelayStateModifier.NOCHANGE.value
            elif state == lcn_defs.MotorStateModifier.TOGGLEONOFF:
                ret += lcn_defs.RelayStateModifier.TOGGLE.value
                ret += lcn_defs.RelayStateModifier.NOCHANGE.value
            elif state == lcn_defs.MotorStateModifier.TOGGLEDIR:
                ret += lcn_defs.RelayStateModifier.NOCHANGE.value
                ret += lcn_defs.RelayStateModifier.TOGGLE.value
            elif state == lcn_defs.MotorStateModifier.CYCLE:
                ret += lcn_defs.RelayStateModifier.TOGGLE.value
                ret += lcn_defs.RelayStateModifier.TOGGLE.value
            elif state == lcn_defs.MotorStateModifier.NOCHANGE:
                ret += lcn_defs.RelayStateModifier.NOCHANGE.value
                ret += lcn_defs.RelayStateModifier.NOCHANGE.value

        return ret

    @staticmethod
    def request_bin_sensors_status():
        """Generates a binary-sensors status request.
         
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return 'SMB'

    @staticmethod
    def var_abs(var, value):
        """Generates a command that sets a variable absolute.

        :param    Var    var:    The target variable to set
        :param    int    value:  The absolute value to set
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        id = lcn_defs.Var.to_set_point_id(var)
        if id != -1:
            # Set absolute (not in PCK yet)
            b1 = id << 6  # 01000000
            b1 |= 0x20  # xx10xxxx (set absolute)
            value -= 1000  # Offset
            b1 |= (value >> 8) & 0x0f  # xxxx1111
            b2 = value & 0xff
            return 'X2{:03d}{:03d}{:03d}'.format(30, b1, b2)

        # Setting variables and thresholds absolute not implemented in LCN firmware yet
        raise ValueError('Wrong variable type.')

    @staticmethod
    def update_status_var(var, value):
        """Generates a command that send variable status updates.
        PCHK provides this variables by itself on selected segments
        is only possible with group 4

        :param    Var    var:    The target variable to set
        :param    int    value:  The absolute value to set
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        id = lcn_defs.Var.to_var_id(var)
        if id != -1:
            # define variable to set, offset 0x01000000
            x2cmd = id | 0x40
            b1 = (value >> 8) & 0xff
            b2 = value & 0xff
            return 'X2{:03d}{:03d}{:03d}'.format(x2cmd, b1, b2)

        # Setting variables and thresholds absolute not implemented in LCN firmware yet
        raise ValueError('Wrong variable type.')

    @staticmethod
    def var_reset(var, is2013=True):
        """Generates a command that resets a variable to 0.

        :param    Var    var:    The target variable to set 0
        :param    bool   is2013: True if the target module's firmware version is 170101 or newer, otherwise False
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        id = lcn_defs.Var.to_var_id(var)
        if id != -1:
            if is2013:
                return 'Z-{:03d}{:04d}'.format(id + 1, 4090)
            else:
                if id == 0:
                    return 'ZS30000'
                else:
                    raise ValueError('Wrong variable type.')

        id = lcn_defs.Var.to_set_point_id(var)
        if id != -1:
            # Set absolute = 0 (not in PCK yet)
            b1 = id << 6  # 01000000
            b1 |= 0x20  # xx10xxxx 9set absolute)
            b2 = 0
            return 'X2{:03d}{:03d}{:03d}'.format(30, b1, b2)

        # Reset for threshold not implemented in LCN firmware yet
        raise ValueError('Wrong variable type.')

    @staticmethod
    def var_rel(var, type, value, is2013=True):
        """Generates a command to change the value of a variable.

        :param    Var          var:    The target variable to change
        :param    RelVarRef    type:   The reference-point
        :param    int          value:  The native LCN value to add/subtract (can be negative)
        :param    bool         is2013: True if the target module's firmware version is 170101 or newer, otherwise False
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        id = lcn_defs.Var.to_var_id(var)
        if id != -1:
            if id == 0:  # Old command for variable 1 / T-var (compatible with all modules)
                return 'Z{:s}{:d}'.format('A' if value >= 0 else 'S', abs(value))
            else:  # New command for variable 1-12 (compatible with all modules, since LCN-PCHK 2.8)
                return 'Z{:s}{:03d}{:d}'.format('+' if value >= 0 else '-', id + 1, abs(value))

        id = lcn_defs.Var.to_set_point_id(var)
        if id != -1:
            return 'RE{:s}S{:s}{:s}{:d}'.format('A' if id == 0 else 'B',
                                                'A' if type == lcn_defs.RelVarRef.CURRENT else 'P',
                                                '+' if value >= 0 else '-',
                                                abs(value))

        register_id = lcn_defs.Var.to_thrs_register_id(var)
        id = lcn_defs.Var.to_thrs_id(var)
        if (register_id != -1) and (id != -1):
            if is2013:  # New command for registers 1-4 (since 170206, LCN-PCHK 2.8)
                return 'SS{:s}{:04d}{:s}R{:d}{:d}'.format('R' if type == lcn_defs.RelVarRef.CURRENT else 'E',
                                                          abs(value),
                                                          'A' if value >= 0 else 'S',
                                                          register_id + 1,
                                                          id + 1)
            elif register_id == 0:  # Old command for register 1 (before 170206)
                return 'SS{:s}{:4d}{:s}{:s}{:s}{:s}{:s}{:s}'.format('R' if type == lcn_defs.RelVarRef.CURRENT else 'E',
                                                                    abs(value),
                                                                    'A' if value >= 0 else 'S',
                                                                    '1' if id == 0 else '0',
                                                                    '1' if id == 1 else '0',
                                                                    '1' if id == 2 else '0',
                                                                    '1' if id == 3 else '0',
                                                                    '1' if id == 4 else '0')

        raise ValueError('Wrong variable type.')

    @staticmethod
    def request_var_status(var, sw_age=0x170206):
        """Generates a variable value request.
    
        :param    Var    var:    The variable to request
        :param    int    swAge:  The target module's firmware version
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if sw_age >= 0x170206:
            id = lcn_defs.Var.to_var_id(var)
            if id != -1:
                return 'MWT{:03d}'.format(id + 1)

            id = lcn_defs.Var.to_set_point_id(var)
            if id != -1:
                return 'MWS{:03d}'.format(id + 1)

            id = lcn_defs.Var.to_thrs_register_id(var)
            if id != -1:
                return 'SE{:03d}'.format(id + 1)  # Whole register

            id = lcn_defs.Var.to_s0_id(var)
            if id != -1:
                return 'MWC{:03d}'.format(id + 1)
        else:
            if var == lcn_defs.Var.VAR1ORTVAR:
                return 'MWV'
            elif var == lcn_defs.Var.VAR2ORR1VAR:
                return 'MWTA'
            elif var == lcn_defs.Var.VAR3ORR2VAR:
                return 'MWTB'
            elif var == lcn_defs.Var.R1VARSETPOINT:
                return 'MWSA'
            elif var == lcn_defs.Var.R2VARSETPOINT:
                return 'MWSB'
            elif var in [lcn_defs.Var.THRS1, lcn_defs.Var.THRS2, lcn_defs.Var.THRS3, lcn_defs.Var.THRS4, lcn_defs.Var.THRS5]:
                return 'SL1'  # Whole register

        raise ValueError('Wrong variable type.')

    @staticmethod
    def request_leds_and_logic_ops():
        """Generates a request for LED and logic-operations states.

        :return: The PCK command (without address header) as text
        :rtype:  str
        """
        return 'SMT'

    @staticmethod
    def control_led(led_id, state):
        """Generates a command to the set the state of a single LED.

        :param    int          led_id:   Led id 0..11
        :param    LedStatus    state:    The state to set
        :return: The PCK command (without address header) as text
        :rtype:  str
        """
        if (led_id < 0) or (led_id > 11):
            raise ValueError('Bad led_id.')
        return 'LA{:03d}{:2}'.format(led_id + 1, state.value)

    @staticmethod
    def send_keys(cmds, keys):
        """Generates a command to send LCN keys.

        :param    SendKeyCOmmand    cmds:    The 4 concrete commands to send for the tables (A-D) as list
        :param    list(bool)        keys:    The tables' 8 key-states (True means "send") as list
        :return:  The PCK command (without address header) as text        
        :rtype:   str
        """
        if (len(cmds) != 4) or (len(keys) != 8):
            raise ValueError('Wrong cmds length or wrong keys length.')
        ret = 'TS'
        for i, cmd in enumerate(cmds):
            if (cmd == lcn_defs.SendKeyCommand.DONTSEND) and (i == 3):
                # By skipping table D (if it is not used), we use the old command
                # for table A-C which is compatible with older LCN modules
                break
            else:
                ret += cmd.value

        for key in keys:
            ret += '1' if key else '0'

        return ret

    @staticmethod
    def send_keys_hit_deferred(table_id, time, time_unit, keys):
        """Generates a command to send LCN keys deferred / delayed.
        
        :param     int         table_id:    Table id 0(A)..3(D)
        :param     int         time:        The delay time
        :param     TimeUnit    time_unit:   The time unit
        :param     list(bool)  keys:        The table's 8 key-states (True means "send") as list
        :return:   The PCK command (without address header) as text        
        :rtype:    str
        """
        if (table_id < 0) or (table_id > 3) or (len(keys) != 8):
            raise ValueError('Bad table_id or keys.')
        ret = 'TV'
        try:
            ret += PckGenerator.TABLE_NAMES[table_id]
        except IndexError:
            raise ValueError('Wrong table_id.')

        ret += '{:03d}'.format(time)
        if time_unit == lcn_defs.TimeUnit.SECONDS:
            if (time < 1) or (time > 60):
                raise ValueError('Wrong time.')
            ret += 'S'
        elif time_unit == lcn_defs.TimeUnit.MINUTES:
            if (time < 1) or (time > 90):
                raise ValueError('Wrong time.')
            ret += 'M'
        elif time_unit == lcn_defs.TimeUnit.HOURS:
            if (time < 1) or (time > 50):
                raise ValueError('Wrong time.')
            ret += 'H'
        elif time_unit == lcn_defs.TimeUnit.DAYS:
            if (time < 1) or (time > 45):
                raise ValueError('Wrong time.')
            ret += 'D'
        else:
            raise ValueError('Wrong time_unit.')

        for key in keys:
            ret += '1' if key else '0'

        return ret

    @staticmethod
    def request_key_lock_status():
        """Generates a request for key-lock states.
        Always requests table A-D. Supported since LCN-PCHK 2.8.

        :return:    The PCK command (without address header) as text
        :rtype:     str
        """
        return 'STX'

    @staticmethod
    def lock_keys(table_id, states):
        """Generates a command to lock keys.

        :param     int           table_id:  Table id 0(A)..3(D)
        :param     list(bool)    states:    The 8 key-lock modifiers as list
        :return:   The PCK command (without address header) as text 
        :rtype:    str
        """
        if (table_id < 0) or (table_id > 3) or (len(states) != 8):
            raise ValueError('Bad table_id or states.')
        try:
            ret = 'TX{:s}'.format(PckGenerator.TABLE_NAMES[table_id])
        except IndexError:
            raise ValueError('Wrong table_id.')

        for state in states:
            ret += state.value

        return ret

    @staticmethod
    def lock_keys_tab_a_temporary(time, time_unit, keys):
        """Generates a command to lock keys for table A temporary.
        There is no hardware-support for locking tables B-D.

        :param     int         time:         The lock time
        :param     TimeUnit    time_unit:    The time unit
        :param     list(bool)  keys:         The 8 key-lock states (True means lock) as list
        :return:   The PCK command (without address header) as text
        :rtype:    str       
        """
        if len(keys) != 8:
            raise ValueError('Wrong keys lenght.')
        ret = 'TXZA{:03d}'.format(time)

        if time_unit == lcn_defs.TimeUnit.SECONDS:
            if (time < 1) or (time > 60):
                raise ValueError('Wrong time.')
            ret += 'S'
        elif time_unit == lcn_defs.TimeUnit.MINUTES:
            if (time < 1) or (time > 90):
                raise ValueError('Wrong time.')
            ret += 'M'
        elif time_unit == lcn_defs.TimeUnit.HOURS:
            if (time < 1) or (time > 50):
                raise ValueError('Wrong time.')
            ret += 'H'
        elif time_unit == lcn_defs.TimeUnit.DAYS:
            if (time < 1) or (time > 45):
                raise ValueError('Wrong time.')
            ret += 'D'
        else:
            raise ValueError('Wrong time_unit.')

        for key in keys:
            ret += '1' if key else '0'

        return ret

    @staticmethod
    def dyn_text_part(row_id, part_id, part):
        """Generates the command header / start for sending dynamic texts.
        Used by LCN-GTxD periphery (supports 4 text rows).
        To complete the command, the text to send must be appended (UTF-8 encoding).
        Texts are split up into up to 5 parts with 12 "UTF-8 bytes" each.

        :param    int    row_id:    Row id 0..3
        :param    int    part_id:   Part id 0..4
        :param    str    part:      Text part (up to 12 bytes), encoded as lcn_defs.LCN_ENCODING
        :return:  The PCK command (without address header) as text       
        :rtype:   str
        """
        if (row_id < 0) or (row_id > 3) or (part_id < 0) or (part_id > 4) or (len(part) > 12):
            raise ValueError('Wrong row_id, part_id or part length.')
        return 'GTDT{:d}{:d}{:s}'.format(row_id + 1, part_id + 1, part.decode(lcn_defs.LCN_ENCODING))

    @staticmethod
    def lock_regulator(reg_id, state):
        """Generates a command to lock a regulator.

        :param    int    reg_id:    Regulator id 0..1
        :param    bool   state:     The lock state
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (reg_id < 0) or (reg_id > 1):
            raise ValueError('Wrong reg_id.')
        return 'RE{:s}X{:s}'.format('A' if reg_id == 0 else 'B',
                                    'S' if state else 'A')



'''
Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
'''

from aenum import Enum, auto
from itertools import product
import math
import re

LCN_ENCODING = 'utf-8'
PATTERN_SPLIT_PORT_PIN = re.compile(r'(?P<port>[a-zA-Z]+)(?P<pin>\d+)')


def split_port_pin(s):
    """Splits the port and the pin from the given input string.
     
    :param    str    s:    Input string
    """
    res = PATTERN_SPLIT_PORT_PIN.findall(s)
    return res[0][0], int(res[0][1])


class OutputPort(Enum):
    """Output port of LCN module.
    """
    OUTPUT1 = 0
    OUTPUT2 = 1
    OUTPUT3 = 2
    OUTPUT4 = 3


class RelayPort(Enum):
    """Relay port of LCN module.
    """
    RELAY1 = 0
    RELAY2 = 1
    RELAY3 = 2
    RELAY4 = 3
    RELAY5 = 4
    RELAY6 = 5
    RELAY7 = 6
    RELAY8 = 7

    MOTORONOFF1 = 0
    MOTORUPDOWN1 = 1
    MOTORONOFF2 = 2
    MOTORUPDOWN2 = 3
    MOTORONOFF3 = 4
    MOTORUPDOWN3 = 5
    MOTORONOFF4 = 6
    MOTORUPDOWN4 = 7


class MotorPort(Enum):
    """Motor ports of LCN module.
    """
    MOTOR1 = 0
    MOTOR2 = 1
    MOTOR3 = 2
    MOTOR4 = 3


class LedPort(Enum):
    """LED port of LCN module.
    """
    LED1 = 0
    LED2 = 1
    LED3 = 2
    LED4 = 3
    LED5 = 4
    LED6 = 5
    LED7 = 6
    LED8 = 7
    LED9 = 8
    LED10 = 9
    LED11 = 10
    LED12 = 11


class LogicOpPort(Enum):
    """Logic Operation port of LCN module
    """
    LOGICOP1 = 0
    LOGICOP2 = 1
    LOGICOP3 = 2
    LOGICOP4 = 3


class BinSensorPort(Enum):
    """Binary sensor port of LCN module.
    """
    BINSENSOR1 = 0
    BINSENSOR2 = 1
    BINSENSOR3 = 2
    BINSENSOR4 = 3
    BINSENSOR5 = 4
    BINSENSOR6 = 5
    BINSENSOR7 = 6
    BINSENSOR8 = 7


# Keys
Key = Enum('Key', ' '.join(['{:s}{:d}'.format(t[0], t[1]) for t in product(['A', 'B', 'C', 'D'], range(1, 9))]), module=__name__)


class OutputPortDimMode(Enum):
    """LCN dimming mode.
    If solely modules with firmware 170206 or newer are present, LCN-PRO automatically programs STEPS200.
    Otherwise the default is STEPS50.
    Since LCN-PCHK doesn't know the current mode, it must explicitly be set.
    """
    STEPS50 = auto()  # 0..50 dimming steps (all LCN module generations)
    STEPS200 = auto()  # 0..200 dimming steps (since 170206)


class OutputPortStatusMode(Enum):
    """
    Tells LCN-PCHK how to format output-port status-messages.
    PERCENT: allows to show the status in half-percent steps (e.g. "10.5").
    NATIVE: is completely backward compatible and there are no restrictions
    concerning the LCN module generations. It requires LCN-PCHK 2.3 or higher though.
    """
    PERCENT = auto()  # Default (compatible with all versions of LCN-PCHK)
    NATIVE = auto()  # 0..200 steps (since LCN-PCHK 2.3)


def time_to_ramp_value(time_msec):
    """
    Converts the given time into an LCN ramp value.
    
    :param    int    time_msec:    The time in milliseconds.
    
    :returns: The (LCN-internal) ramp value (0..250).
    :rtype: int
    """
    if time_msec < 250:
        ret = 0
    elif time_msec < 500:
        ret = 1
    elif time_msec < 660:
        ret = 2
    elif time_msec < 1000:
        ret = 3
    elif time_msec < 1400:
        ret = 4
    elif time_msec < 2000:
        ret = 5
    elif time_msec < 3000:
        ret = 6
    elif time_msec < 4000:
        ret = 7
    elif time_msec < 5000:
        ret = 8
    elif time_msec < 6000:
        ret = 9
    else:
        ret = (time_msec / 1000 - 6) / 2 + 10
        if ret >= 250:
            ret = 250
    return int(ret)


def ramp_value_to_time(ramp_value):
    """Converts the given LCN ramp value into a time.
    
    :param    int    ramp_value:    The LCN ramp value (0..250).
    
    :returns: The ramp time in milliseconds.
    :rtype: int
    """
    if not 0 <= ramp_value <= 250:
        raise ValueError('Ramp value has to be in range 0..250.')

    if ramp_value < 10:
        return [250, 500, 660, 1000, 1400, 2000, 3000, 4000, 5000, 6000][ramp_value]
    else:
        ret = ((ramp_value - 10) * 2 + 6) * 1000
        return int(ret)


class Var(Enum):
    """LCN variable types.
    """
    UNKNOWN = -1  # Used if the real type is not known (yet)
    VAR1ORTVAR = 0
    TVAR = 0
    VAR1 = 0
    VAR2ORR1VAR = 1
    R1VAR = 1
    VAR2 = 1
    VAR3ORR2VAR = 2
    R2VAR = 2
    VAR3 = 2
    VAR4 = 3
    VAR5 = 4
    VAR6 = 5
    VAR7 = 6
    VAR8 = 7
    VAR9 = 8
    VAR10 = 9
    VAR11 = 10
    VAR12 = 11  # Since 170206
    R1VARSETPOINT = auto()
    R2VARSETPOINT = auto()  # Set-points for regulators
    THRS1 = auto()
    THRS2 = auto()
    THRS3 = auto()
    THRS4 = auto()
    THRS5 = auto()  # Register 1 (THRS5 only before 170206)
    THRS2_1 = auto()
    THRS2_2 = auto()
    THRS2_3 = auto()
    THRS2_4 = auto()  # Register 2 (since 2012)
    THRS3_1 = auto()
    THRS3_2 = auto()
    THRS3_3 = auto()
    THRS3_4 = auto()  # Register 3 (since 2012)
    THRS4_1 = auto()
    THRS4_2 = auto()
    THRS4_3 = auto()
    THRS4_4 = auto()  # Register 4 (since 2012)
    S0INPUT1 = auto()
    S0INPUT2 = auto()
    S0INPUT3 = auto()
    S0INPUT4 = auto()  # LCN-BU4LJVarValue

    @staticmethod
    def var_id_to_var(var_id):
        """
        Translates a given id into a variable type.

        :param    int    varId:    The variable id (0..11)
        
        :returns: The translated variable enum.
        :rtype: Var
        """
        if (var_id < 0) or (var_id >= len(Var._var_id_to_var)):
            raise ValueError('Bad var_id.')
        return Var._var_id_to_var[var_id]

    @staticmethod
    def set_point_id_to_var(set_point_id):
        """
        Translates a given id into a LCN set-point variable type.

        :param     int    set_point_id:    Set-point id 0..1
        
        :return: The translated var
        :rtype:  Var     
        """
        if (set_point_id < 0) or (set_point_id >= len(Var._set_point_id_to_var)):
            raise ValueError('Bad set_point_id.')
        return Var._set_point_id_to_var[set_point_id]

    @staticmethod
    def thrs_id_to_var(register_id, thrs_id):
        """
        Translates given ids into a LCN threshold variable type.
        
        :param    int    register_id:    Register id 0..3
        :param    int    thrs_id:        Threshold id 0..4 for register 0, 0..3 for registers 1..3
        
        :return: The translated var
        :rtype:    Var
        """
        if (register_id < 0) or (register_id >= len(Var._thrs_id_to_var)) or (thrs_id < 0) or (thrs_id >= (5 if (register_id == 0) else 4)):
            # print(register_id, thrs_id)
            raise ValueError('Bad register_id and/or thrs_id.')
        return Var._thrs_id_to_var[register_id][thrs_id]

    @staticmethod
    def s0_id_to_var(s0_id):
        """
        Translates a given id into a LCN S0-input variable type.

        :param     int    s0_id:     S0 id 0..3
        
        :return:    The translated var
        :rtype:     Var
        """
        if (s0_id < 0) or (s0_id >= len(Var._s0_id_to_var)):
            raise ValueError('Bad s0_id.')
        return Var._s0_id_to_var_array[s0_id]

    @staticmethod
    def to_var_id(var):
        """
        Translates a given variable type into a variable id.
        
        :param     Var    var:    The variable type to translate
        
        :return:     Variable id 0..11 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.VAR1ORTVAR:
            return 0
        elif var == Var.VAR2ORR1VAR:
            return 1
        elif var == Var.VAR3ORR2VAR:
            return 2
        elif var == Var.VAR4:
            return 3
        elif var == Var.VAR5:
            return 4
        elif var == Var.VAR6:
            return 5
        elif var == Var.VAR7:
            return 6
        elif var == Var.VAR8:
            return 7
        elif var == Var.VAR9:
            return 8
        elif var == Var.VAR10:
            return 9
        elif var == Var.VAR11:
            return 10
        elif var == Var.VAR12:
            return 11
        else:
            return -1

    @staticmethod
    def to_set_point_id(var):
        """
        Translates a given variable type into a set-point id.
        
        :param     Var    var:     The variable type to translate
        
        :return:    Variable id 0..1 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.R1VARSETPOINT:
            return 0
        elif var == Var.R2VARSETPOINT:
            return 1
        else:
            return -1

    @staticmethod
    def to_thrs_register_id(var):
        """
        Translates a given variable type into a threshold register id.

        :param    Var    var:    The variable type to translate
        
        :return:    Register id 0..3 or -1 if wrong type
        :rtype:    int
        """
        if var in [Var.THRS1, Var.THRS2, Var.THRS3, Var.THRS4, Var.THRS5]:
            return 0
        elif var in [Var.THRS2_1, Var.THRS2_2, Var.THRS2_3, Var.THRS2_4]:
            return 1
        elif var in [Var.THRS3_1, Var.THRS3_2, Var.THRS3_3, Var.THRS3_4]:
            return 2
        elif var in [Var.THRS4_1, Var.THRS4_2, Var.THRS4_3, Var.THRS4_4]:
            return 3
        else:
            return -1

    @staticmethod
    def to_thrs_id(var):
        """
        Translates a given variable type into a threshold id.

        :param    Var    var:    The variable type to translate
        
        :return:    Threshold id 0..4 or -1 if wrong type
        :rtype:    int
        """
        if var in [Var.THRS1, Var.THRS2_1, Var.THRS3_1, Var.THRS4_1]:
            return 0
        elif var in [Var.THRS2, Var.THRS2_2, Var.THRS3_2, Var.THRS4_2]:
            return 1
        elif var in [Var.THRS3, Var.THRS2_3, Var.THRS3_3, Var.THRS4_3]:
            return 2
        elif var in [Var.THRS4, Var.THRS2_4, Var.THRS3_4, Var.THRS4_4]:
            return 3
        elif var == Var.THRS5:
            return 4
        else:
            return -1

    @staticmethod
    def to_s0_id(var):
        """
        Translates a given variable type into an S0-input id.

        :param    Var    var:    The variable type to translate
        
        :return:    S0 id 0..3 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.S0INPUT1:
            return 0
        elif var == Var.S0INPUT2:
            return 1
        elif var == Var.S0INPUT3:
            return 2
        elif var == Var.S0INPUT4:
            return 3
        else:
            return -1

    @staticmethod
    def is_lockable_regulator_source(var):
        """
        Checks if the the given variable type is lockable.

        :param    Var    var:    The variable type to check
        
        :return:    True if lockable, otherwise False
        :rtype:    bool
        """
        return var in [Var.R1VARSETPOINT, Var.R2VARSETPOINT]

    @staticmethod
    def use_lcn_special_values(var):
        """
        Checks if the given variable type uses special values.
        Examples for special values: 'No value yet', 'sensor defective' etc.

        :param    Var    var:    The variable type to check
        
        :return:    True if special values are in use, otherwise False
        :rtype:    bool
        """
        return var not in [Var.S0INPU1, Var.S0INPUT2, Var.S0INPUT3, Var.S0INPUT4]

    @staticmethod
    def has_type_in_response(var, sw_age):
        """
        Module-generation check.
        Checks if the given variable type would receive a typed response if
        its status was requested.

        :param    Var    var:    The variable type to check
        :param    int    swAge:  The target LCN-modules firmware version
        
        :return:    True if a response would contain the variable's type, otherwise False
        :rtype:    bool
        """
        if sw_age < 0x170206:
            if var in [Var.VAR1ORTVAR, Var.VAR2ORR1VAR, Var.VAR3ORR2VAR, Var.R1VARSETPOINT, Var.R2VARSETPOINT]:
                return False
        return True

    @staticmethod
    def is_event_based(var, sw_age):
        """
        Module-generation check.
        Checks if the given variable type automatically sends status-updates on
        value-change. It must be polled otherwise.

        :param    Var    var:    The variable type to check
        :param    int    swAge:  The target LCN-module's firmware version
        
        :return:    True if the LCN module supports automatic status-messages for this var, otherwise False
        :rtype:    bool
        """
        if (Var.to_set_point_id(var) != -1) or (Var.to_s0_id(var) != -1):
            return True
        return sw_age >= 0x170206

    @staticmethod
    def should_poll_status_after_command(var, is2013):
        """
        Module-generation check.
        Checks if the target LCN module would automatically send status-updates if
        the given variable type was changed by command.

        :param    Var    var:    The variable type to check
        :param    bool   is2013: The target module's-generation
        
        :return:    True if a poll is required to get the new status-value, otherwise False
        :rtype:    bool
        """
        # Regulator set-points will send status-messages on every change (all firmware versions)
        if Var.to_set_point_id(var) != -1:
            return False

        # Thresholds since 170206 will send status-messages on every change
        if is2013 and (Var.to_thrs_register_id(var != -1)):
            return False

        # Others:
        # - Variables before 170206 will never send any status-messages
        # - Variables since 170206 only send status-messages on "big" changes
        # - Thresholds before 170206 will never send any status-messages
        # - S0-inputs only send status-messages on "big" changes
        # (all "big changes" cases force us to poll the status to get faster updates)
        return True

    @staticmethod
    def should_poll_status_after_regulator_lock(sw_age, lock_state):
        """
        Module-generation check.
        Checks if the target LCN module would automatically send status-updates if
        the given regulator's lock-state was changed by command.

        :param    int       swAge: The target LCN-module's firmware version
        :param    int   lockState: The lock-state sent via command
        
        :return:    True if a poll is required to get the new status-value, otherwise False
        :rtype:    bool
        """
        # LCN modules before 170206 will send an automatic status-message for "lock", but not for "unlock"
        return (lock_state == False) and (sw_age < 0x170206)


# Helper list to get var by numeric id.
Var._var_id_to_var = [Var.VAR1ORTVAR, Var.VAR2ORR1VAR, Var.VAR3ORR2VAR,
                      Var.VAR4, Var.VAR5, Var.VAR6, Var.VAR7, Var.VAR8,
                      Var.VAR9, Var.VAR10, Var.VAR11, Var.VAR12]

# Helper list to get set-point var by numeric id.
Var._set_point_id_to_var = [Var.R1VARSETPOINT, Var.R2VARSETPOINT]

# Helper list to get threshold var by numeric id.
Var._thrs_id_to_var = [[Var.THRS1, Var.THRS2, Var.THRS3, Var.THRS4, Var.THRS5],
                       [Var.THRS2_1, Var.THRS2_2, Var.THRS2_3, Var.THRS2_4],
                       [Var.THRS3_1, Var.THRS3_2, Var.THRS3_3, Var.THRS3_4],
                       [Var.THRS4_1, Var.THRS4_2, Var.THRS4_3, Var.THRS4_4]]

# Helper list to get S0-input var by numeric id.
Var._s0_id_to_var = [Var.S0INPUT1, Var.S0INPUT2, Var.S0INPUT3, Var.S0INPUT4]


class VarUnit(Enum):
    """Measurement units used with LCN variables.
    """
    NATIVE = ''  # LCN internal representation (0 = -100�C for absolute values)
    CELSIUS = '\u00b0C'
    KELVIN = '\u00b0K'
    FAHRENHEIT = '\u00b0F'
    LUX_T = 'Lux_T'
    LUX_I = 'Lux_I'
    METERPERSECOND = 'm/s'  # Used for LCN-WIH wind speed
    PERCENT = '%'  # Used for humidity
    PPM = 'ppm'  # Used by CO2 sensor
    VOLT = 'V'
    AMPERE = 'A'
    DEGREE = '\u00b0'  # Used for angles,

    @staticmethod
    def parse(input):
        input = input.upper()
        if input in ['', 'NATIVE', 'LCN']:
            return VarUnit.NATIVE
        elif input in ['CELSIUS', '\u00b0CELSIUS', '\u00b0C']:
            return VarUnit.CELSIUS
        elif input in ['KELVIN', '\u00b0KELVIN', '\u00b0K']:
            return VarUnit.KELVIN
        elif input in ['FAHRENHEIT', '\u00b0FAHRENHEIT', '\u00b0F']:
            return VarUnit.FAHRENHEIT
        elif input in ['LUX_T', 'LX_T']:
            return VarUnit.LUX_T
        elif input in ['LUX', 'LX']:
            return VarUnit.LUX_I
        elif input == 'M/S':
            return VarUnit.METERPERSECOND
        elif input in ['%', 'PERCENT']:
            return VarUnit.PERCENT
        elif input == 'PPM':
            return VarUnit.PPM
        elif input in ['VOLT', 'V']:
            return VarUnit.VOLT
        elif input in ['AMPERE', 'AMP', 'A']:
            return VarUnit.AMPERE
        elif input in ['DEGREE', '\u00b0']:
            return VarUnit.DEGREE
        else:
            raise ValueError('Bad input.')


class VarValue(object):
    """A value of an LCN variable.

    It internally stores the native LCN value and allows to convert from/into other units.
    Some conversions allow to specify whether the source value is absolute or relative.
    Relative values are used to create varvalues that can be added/subtracted from
    other (absolute) varvalues.    
    
    :param    int    native_value:    The native value       
    """

    def __init__(self, native_value):
        """Constructor with native LCN value.
        """
        self.native_value = native_value

    def is_locked_regulator(self):
        return (self.native_value & 0x8000) != 0

    @staticmethod
    def from_var_unit(v, unit, abs):
        """Creates a variable value from any input.

        :param    float    v:       The input value
        :param    VarUnit  unit:    The input value's unit
        :param    bool     abs:     True for absolute values (relative values are used to add/subtract from other VarValues), otherwise False
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if unit == VarUnit.NATIVE:
            return VarValue.from_native(int(v))
        elif unit == VarUnit.CELSIUS:
            return VarValue.from_celsius(v, abs)
        elif unit == VarUnit.KELVIN:
            return VarValue.from_kelvin(v, abs)
        elif unit == VarUnit.FAHRENHEIT:
            return VarValue.from_fahrenheit(v, abs)
        elif unit == VarUnit.LUX_T:
            return VarValue.from_lux_t(v)
        elif unit == VarUnit.LUX_I:
            return VarValue.from_lux_i(v)
        elif unit == VarUnit.METERPERSECOND:
            return VarValue.from_meters_per_second(v)
        elif unit == VarUnit.PERCENT:
            return VarValue.from_percent(v)
        elif unit == VarUnit.PPM:
            return VarValue.from_ppm(v)
        elif unit == VarUnit.VOLT:
            return VarValue.from_volt(v)
        elif unit == VarUnit.AMPERE:
            return VarValue.from_kelvin(v)
        elif unit == VarUnit.DEGREE:
            return VarValue.from_degree(v)
        else:
            raise ValueError('Wrong unit.')

    @staticmethod
    def from_native(n):
        """Creates a variable value from native input.

        :param    int    n:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(n)

    @staticmethod
    def from_celsius(c, abs=True):
        """Creates a variable value from \u00b0C input.

        :param    float    c:    The input value
        :param    bool     abs:  True for absolute values (relative values are used to add/subtract from other VarValues), otherwise False
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        n = int(round(c * 10))
        return VarValue(n + 1000 if abs else n)

    @staticmethod
    def from_kelvin(k, abs=True):
        """Creates a variable value from \u00b0K input.

        :param    float    k:    The input value
        :param    bool     abs:  True for absolute values (relative values are used to add/subtract from other VarValues), otherwise False
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if abs:
            k -= 273.15

        n = int(round(k * 10))
        return VarValue(n + 1000 if abs else n)

    @staticmethod
    def from_fahrenheit(f, abs=True):
        """Creates a variable value from \u00b0F input.

        :param    float    f:    The input value
        :param    bool     abs:  True for absolute values (relative values are used to add/subtract from other VarValues), otherwise False
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if abs:
            f -= 32

        n = int(round(f / 0.18))
        return VarValue(n + 1000 if abs else n)

    @staticmethod
    def from_lux_t(l):
        """Creates a variable value from lx input.
        Target must be connected to T-port.

        :param    float    l:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(math.log(l) - 1.689646994) / 0.010380664))

    @staticmethod
    def from_lux_i(l):
        """Creates a variable value from lx input.
        Target must be connected to I-port.

        :param    float    l:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(math.log(l) * 100)))

    @staticmethod
    def from_percent(p):
        """Creates a variable value from % input.

        :param    float    p:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(p)))

    @staticmethod
    def from_ppm(p):
        """Creates a variable value from ppm input.
        Used for CO2 sensors.

        :param    float    p:   The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(p)))

    @staticmethod
    def from_meters_per_second(ms):
        """Creates a variable value from m/s input.
        Used for LCN-WIH wind speed.

        :param    float    ms:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(ms * 10)))

    @staticmethod
    def from_volt(v):
        """Creates a variable value from V input.

        :param    float    v:    The input value
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(v * 400)))

    @staticmethod
    def from_ampere(a):
        """Creates a variable value from A input.

        :param    float    a:    The input value
        
        :return: The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(a * 100)))

    @staticmethod
    def from_degree(d, abs=True):
        """Creates a variable value from \u00b0 (angle) input.

        :param    float    d:    The input value
        :param    bool     abs:  True for absolute values (relative values are used to add/subtract from other VarValues), otherwise False
        
        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        n = int(round(d * 10))
        return VarValue(n + 1000 if abs else n)

    def to_var_unit(self, unit, is_lockable_regulator_source=False):
        v = VarValue(self.native_value & 0x7fff if is_lockable_regulator_source else self.native_value)

        if unit == VarUnit.NATIVE:
            return v.to_native()
        elif unit == VarUnit.CELSIUS:
            return v.to_celsius()
        elif unit == VarUnit.KELVIN:
            return v.to_kelvin()
        elif unit == VarUnit.FAHRENHEIT:
            return v.to_fahrenheit()
        elif unit == VarUnit.LUX_T:
            return v.to_lux_t()
        elif unit == VarUnit.LUX_I:
            return v.to_lux_i()
        elif unit == VarUnit.METERPERSECOND:
            return v.to_meters_per_second()
        elif unit == VarUnit.PERCENT:
            return v.to_percent()
        elif unit == VarUnit.PPM:
            return v.to_ppm()
        elif unit == VarUnit.VOLT:
            return v.to_volt()
        elif unit == VarUnit.AMPERE:
            return v.to_ampere()
        elif unit == VarUnit.DEGREE:
            return v.to_degree()
        else:
            raise ValueError('Wrong unit.')

    def to_native(self):
        """Converts to native value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_celsius(self):
        """Converts to \u00b0C value.
        
        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10.

    def to_kelvin(self):
        """Converts to \u00b0K value.
        
        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10. + 273.15

    def to_fahrenheit(self):
        """Converts to \u00b0F value.
        
        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) * 0.18 + 32.

    def to_lux_t(self):
        """Converts to lx value.
        Source must be connected to T-port.
        
        :return:    The converted value
        :rtype:    float
        """
        return math.exp(0.010380664 * self.native_value + 1.689646994)

    def to_lux_i(self):
        """Converts to lx value.
        Source must be connected to I-port.
        
        :return:    The converted value
        :rtype:    float
        """
        return math.exp(self.native_value / 100)

    def to_percent(self):
        """Converts to % value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_ppm(self):
        """Converts to ppm value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_meters_per_sec(self):
        """Converts to m/s value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 10.

    def to_volt(self):
        """
        Converts to V value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 400.

    def to_ampere(self):
        """
        Converts to A value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 100.

    def to_degree(self):
        """
        Converts to \u00b0 value.

        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10.

    def to_var_unit_string(self, unit, is_lockable_regulator_source=False, use_lcn_special_values=False):
        if use_lcn_special_values and (self.native_value == 0xffff):  # No value
            ret = '---'
        elif use_lcn_special_values and ((self.native_value & 0xff00) == 0x8100):  # Undefined
            ret = '---'
        elif use_lcn_special_values and ((self.native_value & 0xff00) == 0x7f00):  # Defective
            ret = '!!!'
        else:
            var = VarValue((self.native_value & 0x7ff) if is_lockable_regulator_source else self.native_value)
            if unit == VarUnit.NATIVE:
                ret = '{:.0f}'.format(var.to_native())
            elif unit == VarUnit.CELSIUS:
                ret = '{:.01f}'.format(var.to_celsius())
            elif unit == VarUnit.KELVIN:
                ret = '{:.01f}'.format(var.to_kelvin())
            elif unit == VarUnit.FAHRENHEIT:
                ret = '{:.01f}'.format(var.to_fahrenheit())
            elif unit == VarUnit.LUX_T:
                if (var.to_native() > 1152):  # Max. value the HW can do
                    ret = '---'
                else:
                    ret = '{:.0f}'.format(var.to_lux_t())
            elif unit == VarUnit.LUX_I:
                if (var.to_native() > 1152):  # Max. value the HW can do
                    ret = '---'
                else:
                    ret = '{:.0f}'.format(var.to_lux_i())
            elif unit == VarUnit.METERPERSECOND:
                ret = '{:.0f}'.format(var.to_meters_per_sec())
            elif unit == VarUnit.PERCENT:
                ret = '{:.0f}'.format(var.to_percent())
            elif unit == VarUnit.PPM:
                ret = '{:.0f}'.format(var.to_ppm())
            elif unit == VarUnit.VOLT:
                ret = '{:.0f}'.format(var.to_volt())
            elif unit == VarUnit.AMPERE:
                ret = '{:.0f}'.format(var.to_ampere())
            elif unit == VarUnit.DEGREE:
                ret = '{:.0f}'.format(var.to_degree())
            else:
                raise ValueError('Wrong unit.')

        # handle locked regulators
        if is_lockable_regulator_source and self.is_locked_regulator():
            ret = '({:s})'.format(ret)

        return ret


class LedStatus(Enum):
    """Possible states for LCN LEDs
    """
    OFF = 'A'
    ON = 'E'
    BLINK = 'B'
    FLICKER = 'F'


class LogicOpStatus(Enum):
    """Possible states for LCN logic-operations.
    """
    NOT = 'N'
    OR = "T"  # Note: Actually not correct since AND won't be OR also
    AND = 'V'


class TimeUnit(Enum):
    """Time units used for several LCN commands.
    """
    SECONDS = 'S'
    MINUTES = 'M'
    HOURS = 'H'
    DAYS = 'D'

    @staticmethod
    def parse(input):
        """Parses the given input into a time unit.
        It supports several alternative terms.

        :param    str    input:    The text to parse
        :return:    TimeUnit enum
        :rtype:    TimeUnit
        """
        input = input.upper()
        if input in ['SECONDS', 'SECOND', 'SEC', 'S']:
            return TimeUnit.SECONDS
        elif input in ['MINUTES', 'MINUTE', 'MIN', 'M']:
            return TimeUnit.MINUTES
        elif input in ['HOURS', 'HOUR', 'H']:
            return TimeUnit.HOURS
        elif input in ['DAYS', 'DAY', 'D']:
            return TimeUnit.DAYS
        else:
            raise ValueError('Bad time unit input.')


class RelayStateModifier(Enum):
    """Relay-state modifiers used in LCN commands.
    """
    ON = '1'
    OFF = '0'
    TOGGLE = 'U'
    NOCHANGE = '-'


class MotorStateModifier(Enum):
    """Motor-state modifiers used in LCN commands.
    LCN module has to be configured for motors connected to relays.
    """
    UP = 'U'
    DOWN = 'D'
    STOP = 'S'
    TOGGLEONOFF = 'T'  # toggle on/off
    TOGGLEDIR = 'D'  # toggle direction
    CYCLE = 'C'  # up, stop, down, stop, ...
    NOCHANGE = '-'


class RelVarRef(Enum):
    """Value-reference for relative LCN variable commands.
    """
    CURRENT = auto()
    PROG = auto()  # Programmed value (LCN-PRO). Relevant for set-points and thresholds.


class SendKeyCommand(Enum):
    """Command types used when sending LCN keys.
    """
    HIT = 'K'
    MAKE = 'L'
    BREAK = 'O'
    DONTSEND = '-'


class KeyLockStateModifier(Enum):
    """Key-lock modifiers used in LCN commands.
    """
    ON = '1'
    OFF = '0'
    TOGGLE = 'U'
    NOCHANGE = '-'


default_connection_settings = {'NUM_TRIES': 3,  # Total number of request to sent before going into failed-state.
                               'SK_NUM_TRIES': 3,  # Total number of segment coupler scan tries
                               'DIM_MODE': OutputPortDimMode.STEPS50,
                               'PING_TIMEOUT': 600000,  # The default timeout for pings sent to PCHK.
                               'MAX_STATUS_EVENTBASED_VALUEAGE_MSEC': 600000,  # Poll interval for status values that automatically send their values on change.
                               'MAX_STATUS_POLLED_VALUEAGE_MSEC': 30000,  # Poll interval for status values that do not send their values on change (always polled).
                               'STATUS_REQUEST_DELAY_AFTER_COMMAND_MSEC': 2000  # Status request delay after a command has been send which potentially changed that status.
                               }



'''
Copyright (c) 2006-2018 by the respective copyright holders.

All rights reserved. This program and the accompanying materials
are made available under the terms of the Eclipse Public License v1.0
which accompanies this distribution, and is available at
http://www.eclipse.org/legal/epl-v10.html

Contributors:
  Andre Lengwenus - port to Python and further improvements
  Tobias Juettner - initial LCN binding for openHAB (Java)
'''

import logging

_LOGGER = logging.getLogger(__name__)

# The default timeout to use for requests. Worst case: Requesting threshold 4-4 takes at least 1.8s
DEFAULT_TIMEOUT_MSEC = 3500


class TimeoutRetryHandler(object):
    """Manages timeout and retry logic for an LCN request.
    
    :param           loop:           The asyncio event loop
    :param    int    num_tries:      The maximum number of tries until the request is marked as failed (-1 means forever)
    :param    int    timeout_msec:   Default timeout in milliseconds
    
    """

    def __init__(self, loop, num_tries=3, timeout_msec=DEFAULT_TIMEOUT_MSEC):
        """Constructor.
        """
        self.loop = loop
        self.num_tries = num_tries
        self._timeout_msec = timeout_msec

        self._timeout_handle = None

        self.reset()

    def is_active(self):
        """Checks whether the request logic is active.
        """
        return self._timeout_handle is not None

    def set_timeout_msec(self, timeout_msec):
        self._timeout_msec = timeout_msec

    def set_timeout_callback(self, func):
        """Timeout_callback function is called, if timeout expires.
        Function has to take one argument:
        Returns failed state (True if failed)
        """
        self._timeout_callback = func

    def on_timeout(self):
        if self._timeout_callback is not None:
            self._timeout_callback(self._num_tries_left == 0)

        if self._num_tries_left == 0:
            self.cancel()
            return
        elif self._num_tries_left > 0 :
            self._num_tries_left -= 1
        self._timeout_handle = self.loop.call_later(self._timeout_msec / 1000, self.on_timeout)

    def reset(self):
        if self._timeout_handle is not None:
            self._timeout_handle.cancel()
            self._timeout_handle = None
        self._num_tries_left = 0

    def activate(self, timeout_callback=None):
        """Schedules the next request.
        """
#         if self.is_active():
#             return
        self.reset()
        self._num_tries_left = self.num_tries
        if timeout_callback is not None:
            self.set_timeout_callback(timeout_callback)
        self._timeout_handle = self.loop.call_soon(self.on_timeout)

    def cancel(self):
        """Must be called when a response (requested or not) has been received.
        """
        self.reset()
